@@ -6,6 +6,7 @@
 #include <sys/time.h>
 
 #include <amqp.h>
+#include <amqp_tcp_socket.h>
 
 #include "connection.h"
 #include "distmeta.h"
@@ -40,12 +41,18 @@ AMQTable_SetStringValue(amqp_table_t*, amqp_bytes_t, amqp_bytes_t);
 _PYRMQ_INLINE void
 AMQTable_SetIntValue(amqp_table_t *, amqp_bytes_t, int);
 
+_PYRMQ_INLINE void
+AMQTable_SetNilValue(amqp_table_t *, amqp_bytes_t);
+
 _PYRMQ_INLINE void
 AMQTable_SetDoubleValue(amqp_table_t *, amqp_bytes_t, double);
 
 _PYRMQ_INLINE amqp_field_value_t*
 AMQArray_AddEntry(amqp_array_t*);
 
+_PYRMQ_INLINE void
+AMQArray_SetNilValue(amqp_array_t*);
+
 _PYRMQ_INLINE void
 AMQArray_SetTableValue(amqp_array_t*, amqp_table_t);
 
@@ -68,7 +75,8 @@ basic_properties_to_PyDict(amqp_basic_properties_t*, PyObject*);
 _PYRMQ_INLINE int
 PyDict_to_basic_properties(PyObject *,
                            amqp_basic_properties_t *,
-                           amqp_connection_state_t );
+                           amqp_connection_state_t,
+                           amqp_pool_t *);
 
 _PYRMQ_INLINE void
 amqp_basic_deliver_to_PyDict(PyObject *, uint64_t, amqp_bytes_t,
@@ -96,12 +104,13 @@ int PyRabbitMQ_HandleAMQError(PyRabbitMQ_Connection *, unsigned int,
 void PyRabbitMQ_SetErr_UnexpectedHeader(amqp_frame_t*);
 int PyRabbitMQ_Not_Connected(PyRabbitMQ_Connection *);
 
-static amqp_table_t PyDict_ToAMQTable(amqp_connection_state_t, PyObject *);
-static amqp_array_t PyList_ToAMQArray(amqp_connection_state_t, PyObject *);
+static amqp_table_t PyDict_ToAMQTable(amqp_connection_state_t, PyObject *, amqp_pool_t *);
+static amqp_array_t PyIter_ToAMQArray(amqp_connection_state_t, PyObject *, amqp_pool_t *);
 
 static PyObject* AMQTable_toPyDict(amqp_table_t *table);
 static PyObject* AMQArray_toPyList(amqp_array_t *array);
 
+int PyRabbitMQ_HandleAMQStatus(int, const char *);
 
 int PyRabbitMQ_Not_Connected(PyRabbitMQ_Connection *self)
 {
@@ -160,6 +169,12 @@ AMQTable_SetIntValue(amqp_table_t *table,
     entry->value.value.i32 = value;
 }
 
+_PYRMQ_INLINE void
+AMQTable_SetNilValue(amqp_table_t *table, amqp_bytes_t key) {
+    amqp_table_entry_t *entry = AMQTable_AddEntry(table, key);
+    entry->value.kind = AMQP_FIELD_KIND_VOID;
+}
+
 _PYRMQ_INLINE void
 AMQTable_SetDoubleValue(amqp_table_t *table,
                         amqp_bytes_t key, double value)
@@ -177,6 +192,13 @@ AMQArray_AddEntry(amqp_array_t *array)
     return entry;
 }
 
+_PYRMQ_INLINE void
+AMQArray_SetNilValue(amqp_array_t *array)
+{
+    amqp_field_value_t *entry = AMQArray_AddEntry(array);
+    entry->kind = AMQP_FIELD_KIND_VOID;
+}
+
 _PYRMQ_INLINE void
 AMQArray_SetTableValue(amqp_array_t *array, amqp_table_t value)
 {
@@ -210,7 +232,7 @@ AMQArray_SetIntValue(amqp_array_t *array, int value)
 }
 
 static amqp_table_t
-PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
+PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src, amqp_pool_t *pool)
 {
     PyObject *dkey = NULL;
     PyObject *dvalue = NULL;
@@ -225,21 +247,24 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
 
     /* allocate new table */
     dst.num_entries = 0;
-    dst.entries = amqp_pool_alloc(&conn->frame_pool,
-                           size * sizeof(amqp_table_entry_t));
+    dst.entries = amqp_pool_alloc(pool, size * sizeof(amqp_table_entry_t));
     while (PyDict_Next(src, &pos, &dkey, &dvalue)) {
 
-        if (PyDict_Check(dvalue)) {
+        if (dvalue == Py_None) {
+            /* None */
+            AMQTable_SetNilValue(&dst, PyString_AS_AMQBYTES(dkey));
+        }
+        else if (PyDict_Check(dvalue)) {
             /* Dict */
             AMQTable_SetTableValue(&dst,
                     PyString_AS_AMQBYTES(dkey),
-                    PyDict_ToAMQTable(conn, dvalue));
+                    PyDict_ToAMQTable(conn, dvalue, pool));
         }
-        else if (PyList_Check(dvalue)) {
+        else if (PyList_Check(dvalue) || PyTuple_Check(dvalue)) {
             /* List */
             AMQTable_SetArrayValue(&dst,
                     PyString_AS_AMQBYTES(dkey),
-                    PyList_ToAMQArray(conn, dvalue));
+                    PyIter_ToAMQArray(conn, dvalue, pool));
         }
         else if (PyLong_Check(dvalue) || PyInt_Check(dvalue)) {
             /* Int | Long */
@@ -251,6 +276,15 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
                     clong_value
             );
         }
+        else if (PyFloat_Check(dvalue)) {
+            cdouble_value = PyFloat_AsDouble(dvalue);
+            if (PyErr_Occurred())
+                goto error;
+            AMQTable_SetDoubleValue(&dst,
+                PyString_AS_AMQBYTES(dkey),
+                    cdouble_value
+            );
+        }
         else {
             /* String | Unicode */
             is_unicode = PyUnicode_Check(dvalue);
@@ -265,23 +299,11 @@ PyDict_ToAMQTable(amqp_connection_state_t conn, PyObject *src)
                 );
             }
             else {
-                cdouble_value = PyFloat_AsDouble(dvalue);
-                if (PyErr_Occurred())
-                    goto error;
-                if (PyFloat_Check(dvalue)) {
-
-                    AMQTable_SetDoubleValue(&dst,
-                        PyString_AS_AMQBYTES(dkey),
-                        cdouble_value
-                    );
-                }
-                else {
-                    /* unsupported type */
-                    PyErr_Format(PyExc_ValueError,
-                        "Table member %s is of an unsupported type",
-                        PyString_AS_STRING(dkey));
-                    goto error;
-            }
+                /* unsupported type */
+                PyErr_Format(PyExc_ValueError,
+                    "Table member %s is of an unsupported type",
+                    PyString_AS_STRING(dkey));
+                goto error;
             }
         }
     }
@@ -292,64 +314,73 @@ error:
 }
 
 static amqp_array_t
-PyList_ToAMQArray(amqp_connection_state_t conn, PyObject *src)
+PyIter_ToAMQArray(amqp_connection_state_t conn, PyObject *src, amqp_pool_t *pool)
 {
-    PyObject *dvalue = NULL;
-    PY_SIZE_TYPE size = 0;
     PY_SIZE_TYPE pos = 0;
     uint64_t clong_value = 0;
     int is_unicode = 0;
     amqp_array_t dst = AMQP_EMPTY_ARRAY;
 
-    size = PyList_Size(src);
+    PY_SIZE_TYPE size = PySequence_Size(src);
+    if (size == -1) return dst;
 
-    /* allocate new array */
+    PyObject *iterator = PyObject_GetIter(src);
+    if (iterator == NULL) return dst;
+    PyObject *item = NULL;
+
+    /* allocate new amqp array */
     dst.num_entries = 0;
-    dst.entries = amqp_pool_alloc(&conn->frame_pool,
-                           size * sizeof(amqp_field_value_t));
-    for (pos = 0; pos < size; ++pos) {
+    dst.entries = amqp_pool_alloc(pool, size * sizeof(amqp_field_value_t));
 
-        dvalue = PyList_GetItem(src, pos);
-
-        if (PyDict_Check(dvalue)) {
+    while (item = PyIter_Next(iterator)) {
+        if (item == Py_None) {
+            /* None */
+            AMQArray_SetNilValue(&dst);
+        }
+        else if (PyDict_Check(item)) {
             /* Dict */
-            AMQArray_SetTableValue(&dst,
-                    PyDict_ToAMQTable(conn, dvalue));
+            AMQArray_SetTableValue(
+                &dst, PyDict_ToAMQTable(conn, item, pool));
         }
-        else if (PyList_Check(dvalue)) {
+        else if (PyList_Check(item) || PyTuple_Check(item)) {
             /* List */
-            AMQArray_SetArrayValue(&dst,
-                    PyList_ToAMQArray(conn, dvalue));
+            AMQArray_SetArrayValue(
+                &dst, PyIter_ToAMQArray(conn, item, pool));
         }
-        else if (PyLong_Check(dvalue) || PyInt_Check(dvalue)) {
+        else if (PyLong_Check(item) || PyInt_Check(item)) {
             /* Int | Long */
-            clong_value = (int64_t)PyLong_AsLong(dvalue);
+            clong_value = (int64_t)PyLong_AsLong(item);
             AMQArray_SetIntValue(&dst, clong_value);
         }
         else {
             /* String | Unicode */
-            is_unicode = PyUnicode_Check(dvalue);
-            if (is_unicode || PyString_Check(dvalue)) {
+            is_unicode = PyUnicode_Check(item);
+            if (is_unicode || PyString_Check(item)) {
                 if (is_unicode) {
-                    if ((dvalue = PyUnicode_AsASCIIString(dvalue)) == NULL)
-                        goto error;
+                    if ((item = PyUnicode_AsASCIIString(item)) == NULL)
+                        goto item_error;
                 }
-                AMQArray_SetStringValue(&dst,
-                        PyString_AS_AMQBYTES(dvalue));
+                AMQArray_SetStringValue(
+                    &dst, PyString_AS_AMQBYTES(item));
             }
             else {
                 /* unsupported type */
                 PyErr_Format(PyExc_ValueError,
                     "Array member at index %lu, %s, is of an unsupported type",
-                    pos, PyObject_REPR(dvalue));
-                goto error;
+                    pos, PyObject_REPR(item));
+                goto item_error;
             }
         }
+        Py_XDECREF(item);
     }
+
     return dst;
+item_error:
+    Py_XDECREF(item);
 error:
+    Py_XDECREF(iterator);
     assert(PyErr_Occurred());
-    return AMQP_EMPTY_ARRAY;
+    return dst;
 }
 
 
@@ -480,6 +511,9 @@ AMQTable_toPyDict(amqp_table_t *table)
         int i;
         for (i = 0; i < table->num_entries; ++i, key=value=NULL) {
             switch (table->entries[i].value.kind) {
+                case AMQP_FIELD_KIND_VOID:
+                    value = Py_None;
+                    break;
                 case AMQP_FIELD_KIND_BOOLEAN:
                     value = PyBool_FromLong(AMQTable_VAL(table, i, boolean));
                     break;
@@ -530,11 +564,11 @@ AMQTable_toPyDict(amqp_table_t *table)
             }
 
             key = AMQTable_TO_PYKEY(table, i);
-            if (value)
-                PyDICT_SETKV_DECREF(dict, key, value);
-            else
+            if (value == Py_None || value == NULL)
                 /* unsupported type */
                 PyDICT_SETNONE_DECREF(dict, key);
+            else
+                PyDICT_SETKV_DECREF(dict, key, value);
 
         }
     }
@@ -613,11 +647,11 @@ AMQArray_toPyList(amqp_array_t *array)
     return list;
 }
 
-
 _PYRMQ_INLINE int
 PyDict_to_basic_properties(PyObject *p,
                            amqp_basic_properties_t *props,
-                           amqp_connection_state_t conn)
+                           amqp_connection_state_t conn,
+                           amqp_pool_t *pool)
 {
     PyObject *value = NULL;
     props->headers = AMQP_EMPTY_TABLE;
@@ -682,9 +716,8 @@ PyDict_to_basic_properties(PyObject *p,
     }
 
     if ((value = PyDict_GetItemString(p, "headers")) != NULL) {
-        props->headers = PyDict_ToAMQTable(conn, value);
-        if (PyErr_Occurred())
-            return -1;
+        props->headers = PyDict_ToAMQTable(conn, value, pool);
+        if (PyErr_Occurred()) return -1;
     }
     return 1;
 }
@@ -743,6 +776,18 @@ PyRabbitMQ_HandlePollError(int ready)
 }
 
 
+int PyRabbitMQ_HandleAMQStatus(int status, const char *context)
+{
+    char errorstr[1024];
+    if (status) {
+        snprintf(errorstr, sizeof(errorstr), "%s: %s",
+                context, amqp_error_string2(status));
+        PyErr_SetString(PyRabbitMQExc_ConnectionError, errorstr);
+    }
+    return status;
+}
+
+
 int PyRabbitMQ_HandleAMQError(PyRabbitMQ_Connection *self, unsigned int channel,
         amqp_rpc_reply_t reply, const char *context)
 {
@@ -761,7 +806,7 @@ int PyRabbitMQ_HandleAMQError(PyRabbitMQ_Connection *self, unsigned int channel,
             snprintf(errorstr, sizeof(errorstr), "%s: %s",
                     context,
                     reply.library_error
-                        ? amqp_error_string(reply.library_error)
+                        ? amqp_error_string2(reply.library_error)
                         : "(end-of-stream)");
             goto connerror;
 
@@ -848,6 +893,7 @@ PyRabbitMQ_ConnectionType_new(PyTypeObject *type,
         self->port = 5672;
         self->sockfd = 0;
         self->connected = 0;
+        self->server_properties = NULL;
         self->callbacks = NULL;
     }
     return self;
@@ -863,6 +909,7 @@ PyRabbitMQ_ConnectionType_dealloc(PyRabbitMQ_Connection *self)
     if (self->weakreflist != NULL)
         PyObject_ClearWeakRefs((PyObject*)self);
     Py_XDECREF(self->callbacks);
+    Py_XDECREF(self->server_properties);
     self->ob_type->tp_free(self);
 }
 
@@ -910,8 +957,8 @@ PyRabbitMQ_ConnectionType_init(PyRabbitMQ_Connection *self,
     self->heartbeat = heartbeat;
     self->weakreflist = NULL;
     self->callbacks = PyDict_New();
-    if (self->callbacks == NULL)
-        return -1;
+    if (self->callbacks == NULL) return -1;
+    self->server_properties = NULL;
 
     return 0;
 }
@@ -923,7 +970,13 @@ PyRabbitMQ_ConnectionType_init(PyRabbitMQ_Connection *self,
 static PyObject*
 PyRabbitMQ_Connection_fileno(PyRabbitMQ_Connection *self)
 {
-    return PyInt_FromLong((long)self->sockfd);
+    if (self->sockfd > 0) {
+        return PyInt_FromLong((long)self->sockfd);
+    }
+    else {
+        PyErr_SetString(PyExc_ValueError, "Socket not connected");
+        return 0;
+    }
 }
 
 
@@ -933,6 +986,8 @@ PyRabbitMQ_Connection_fileno(PyRabbitMQ_Connection *self)
 static PyObject*
 PyRabbitMQ_Connection_connect(PyRabbitMQ_Connection *self)
 {
+    int status;
+    amqp_socket_t *socket = NULL;
     amqp_rpc_reply_t reply;
 
     if (self->connected) {
@@ -941,13 +996,22 @@ PyRabbitMQ_Connection_connect(PyRabbitMQ_Connection *self)
     }
     Py_BEGIN_ALLOW_THREADS;
     self->conn = amqp_new_connection();
-    self->sockfd = amqp_open_socket(self->hostname, self->port);
+    socket = amqp_tcp_socket_new(self->conn);
     Py_END_ALLOW_THREADS;
-    if (!PyRabbitMQ_HandleError(self->sockfd, "Error opening socket"))
+
+    if (!socket) {
+        PyErr_NoMemory();
         goto error;
+    }
+    Py_BEGIN_ALLOW_THREADS;
+    status = amqp_socket_open(socket, self->hostname, self->port);
+    Py_END_ALLOW_THREADS;
+    if (PyRabbitMQ_HandleAMQStatus(status, "Error opening socket")) {
+        goto error;
+    }
 
     Py_BEGIN_ALLOW_THREADS;
-    amqp_set_sockfd(self->conn, self->sockfd);
+    self->sockfd = amqp_socket_get_sockfd(socket);
     reply = amqp_login(self->conn, self->virtual_host, self->channel_max,
                        self->frame_max, self->heartbeat,
                        AMQP_SASL_METHOD_PLAIN, self->userid, self->password);
@@ -960,6 +1024,7 @@ PyRabbitMQ_Connection_connect(PyRabbitMQ_Connection *self)
     self->channel_max = self->conn->channel_max;
     self->frame_max = self->conn->frame_max;
     self->heartbeat = self->conn->heartbeat;
+    self->server_properties = AMQTable_toPyDict(amqp_get_server_properties(self->conn));
     Py_RETURN_NONE;
 error:
     PyRabbitMQ_Connection_close(self);
@@ -982,7 +1047,7 @@ PyRabbitMQ_Connection_close(PyRabbitMQ_Connection *self)
         Py_BEGIN_ALLOW_THREADS
         reply = amqp_connection_close(self->conn, AMQP_REPLY_SUCCESS);
         amqp_destroy_connection(self->conn);
-        close(self->sockfd);
+        self->sockfd = 0;
         Py_END_ALLOW_THREADS
     }
 
@@ -1038,7 +1103,7 @@ PyRabbitMQ_Connection_destroy_channel(PyRabbitMQ_Connection *self,
     amqp_rpc_reply_t reply;
     Py_BEGIN_ALLOW_THREADS;
     reply = amqp_channel_close(self->conn, channel, AMQP_REPLY_SUCCESS);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     return PyRabbitMQ_HandleAMQError(self, channel, reply, "Couldn't close channel");
@@ -1126,6 +1191,40 @@ finally:
     return retval;
 }
 
+void
+PyRabbitMQ_SetErr_ReceivedFrame(PyRabbitMQ_Connection *self, amqp_frame_t* frame)
+{
+    static char errstr[512];
+
+    switch(frame->payload.method.id) {
+        case AMQP_CHANNEL_CLOSE_METHOD: {
+            amqp_channel_close_t *chanm = (amqp_channel_close_t *)frame->payload.method.decoded;
+            snprintf(errstr, sizeof(errstr),
+                     "channel error %d, message: %.*s",
+                     chanm->reply_code,
+                     (int) chanm->reply_text.len,
+                     (char *) chanm->reply_text.bytes);
+            PyErr_SetString(PyRabbitMQExc_ChannelError, errstr);
+            PyRabbitMQ_revive_channel(self, frame->channel);
+            break;
+        }
+        case AMQP_CONNECTION_CLOSE_METHOD: {
+            amqp_connection_close_t *connm = (amqp_connection_close_t *)frame->payload.method.decoded;
+            snprintf(errstr, sizeof(errstr),
+                "server connection error %d message: %.*s",
+                connm->reply_code,
+                (int) connm->reply_text.len,
+                (char *) connm->reply_text.bytes);
+            PyErr_SetString(PyRabbitMQExc_ConnectionError, errstr);
+            PyRabbitMQ_Connection_close(self);
+            break;
+        }
+        default: {
+            PyErr_SetString(PyRabbitMQExc_ConnectionError, "Bad frame read");
+            break;
+        }
+    }
+}
 
 int
 PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
@@ -1157,9 +1256,8 @@ PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
             retval = amqp_simple_wait_frame(conn, &frame);
             Py_END_ALLOW_THREADS;
             if (retval < 0) break;
-            if (frame.frame_type != AMQP_FRAME_METHOD
-                    || frame.payload.method.id != AMQP_BASIC_DELIVER_METHOD)
-                continue;
+            if (frame.frame_type != AMQP_FRAME_METHOD) continue;
+            if (frame.payload.method.id != AMQP_BASIC_DELIVER_METHOD) goto altframe;
 
             delivery_info = PyDict_New();
             deliver = (amqp_basic_deliver_t *)frame.payload.method.decoded;
@@ -1183,6 +1281,7 @@ PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
         Py_END_ALLOW_THREADS;
         if (retval < 0) break;
 
+        if (frame.frame_type == AMQP_FRAME_METHOD) goto altframe;
         if (frame.frame_type != AMQP_FRAME_HEADER) {
             PyRabbitMQ_SetErr_UnexpectedHeader(&frame);
             goto finally;
@@ -1205,6 +1304,7 @@ PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
             Py_END_ALLOW_THREADS;
             if (retval < 0) break;
 
+            if (frame.frame_type == AMQP_FRAME_METHOD) goto altframe;
             if (frame.frame_type != AMQP_FRAME_BODY) {
                 PyErr_SetString(PyRabbitMQExc_ChannelError,
                     "Expected body, got unexpected frame");
@@ -1254,6 +1354,8 @@ PyRabbitMQ_recv(PyRabbitMQ_Connection *self, PyObject *p,
     }
     goto finally;
 
+altframe:
+    PyRabbitMQ_SetErr_ReceivedFrame(self, &frame);
 error:
     retval = -1;
 
@@ -1281,6 +1383,7 @@ PyRabbitMQ_Connection_queue_bind(PyRabbitMQ_Connection *self,
     PyObject *arguments = NULL;
 
     amqp_table_t bargs = AMQP_EMPTY_TABLE;
+    amqp_pool_t *channel_pool = NULL;
     amqp_rpc_reply_t reply;
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1293,7 +1396,12 @@ PyRabbitMQ_Connection_queue_bind(PyRabbitMQ_Connection *self,
     if ((exchange = Maybe_Unicode(exchange)) == NULL) goto bail;
     if ((routing_key = Maybe_Unicode(routing_key)) == NULL) goto bail;
 
-    bargs = PyDict_ToAMQTable(self->conn, arguments);
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    bargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred())
         goto bail;
 
@@ -1304,7 +1412,7 @@ PyRabbitMQ_Connection_queue_bind(PyRabbitMQ_Connection *self,
                         PyString_AS_AMQBYTES(routing_key),
                         bargs);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.bind"))
@@ -1330,6 +1438,7 @@ PyRabbitMQ_Connection_queue_unbind(PyRabbitMQ_Connection *self,
     PyObject *arguments = NULL;
 
     amqp_table_t uargs = AMQP_EMPTY_TABLE;
+    amqp_pool_t *channel_pool = NULL;
     amqp_rpc_reply_t reply;
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1341,7 +1450,13 @@ PyRabbitMQ_Connection_queue_unbind(PyRabbitMQ_Connection *self,
     if ((queue = Maybe_Unicode(queue)) == NULL) goto bail;
     if ((exchange = Maybe_Unicode(exchange)) == NULL) goto bail;
     if ((routing_key = Maybe_Unicode(routing_key)) == NULL) goto bail;
-    uargs = PyDict_ToAMQTable(self->conn, arguments);
+
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    uargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred())
         goto bail;
 
@@ -1352,7 +1467,7 @@ PyRabbitMQ_Connection_queue_unbind(PyRabbitMQ_Connection *self,
                       PyString_AS_AMQBYTES(routing_key),
                       uargs);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.unbind"))
@@ -1393,7 +1508,7 @@ PyRabbitMQ_Connection_queue_delete(PyRabbitMQ_Connection *self,
             (amqp_boolean_t)if_empty);
     if (ok == NULL)
         reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (ok == NULL && PyRabbitMQ_HandleAMQError(self, channel,
@@ -1423,6 +1538,7 @@ PyRabbitMQ_Connection_queue_declare(PyRabbitMQ_Connection *self,
 
     amqp_queue_declare_ok_t *ok;
     amqp_rpc_reply_t reply;
+    amqp_pool_t *channel_pool = NULL;
     amqp_table_t qargs = AMQP_EMPTY_TABLE;
     PyObject *ret = NULL;
 
@@ -1434,7 +1550,12 @@ PyRabbitMQ_Connection_queue_declare(PyRabbitMQ_Connection *self,
             &exclusive, &auto_delete, &arguments))
         goto bail;
     if ((queue = Maybe_Unicode(queue)) == NULL) goto bail;
-    qargs = PyDict_ToAMQTable(self->conn, arguments);
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    qargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred())
         goto bail;
 
@@ -1489,7 +1610,7 @@ PyRabbitMQ_Connection_queue_purge(PyRabbitMQ_Connection *self,
     ok = amqp_queue_purge(self->conn, channel,
                           PyString_AS_AMQBYTES(queue));
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "queue.purge"))
@@ -1519,6 +1640,7 @@ PyRabbitMQ_Connection_exchange_declare(PyRabbitMQ_Connection *self,
     unsigned int auto_delete = 0;
 
     amqp_table_t eargs = AMQP_EMPTY_TABLE;
+    amqp_pool_t *channel_pool = NULL;
     amqp_rpc_reply_t reply;
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1530,7 +1652,12 @@ PyRabbitMQ_Connection_exchange_declare(PyRabbitMQ_Connection *self,
         goto bail;
     if ((exchange = Maybe_Unicode(exchange)) == NULL) goto bail;
     if ((type = Maybe_Unicode(type)) == NULL) goto bail;
-    eargs = PyDict_ToAMQTable(self->conn, arguments);
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    eargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred())
         goto bail;
 
@@ -1577,7 +1704,7 @@ PyRabbitMQ_Connection_exchange_delete(PyRabbitMQ_Connection *self,
                          PyString_AS_AMQBYTES(exchange),
                          (amqp_boolean_t)if_unused);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "exchange.delete"))
@@ -1608,6 +1735,7 @@ PyRabbitMQ_Connection_basic_publish(PyRabbitMQ_Connection *self,
     int ret = 0;
     amqp_basic_properties_t props;
     amqp_bytes_t bytes;
+    amqp_pool_t *channel_pool = NULL;
     memset(&props, 0, sizeof(props));
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1621,8 +1749,10 @@ PyRabbitMQ_Connection_basic_publish(PyRabbitMQ_Connection *self,
     if ((routing_key = Maybe_Unicode(routing_key)) == NULL) goto bail;
 
     Py_INCREF(propdict);
-    if (!PyDict_to_basic_properties(propdict, &props, self->conn))
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (PyDict_to_basic_properties(propdict, &props, self->conn, channel_pool) < 1) {
         goto bail;
+    }
     Py_DECREF(propdict);
 
     bytes.len = (size_t)body_size;
@@ -1636,10 +1766,12 @@ PyRabbitMQ_Connection_basic_publish(PyRabbitMQ_Connection *self,
                              (amqp_boolean_t)immediate,
                              &props,
                              bytes);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
-    if (!PyRabbitMQ_HandleError(ret, "basic.publish"))
+    if (!PyRabbitMQ_HandleError(ret, "basic.publish")) {
         goto error;
+    }
     Py_RETURN_NONE;
 
 error:
@@ -1742,7 +1874,7 @@ PyRabbitMQ_Connection_basic_cancel(PyRabbitMQ_Connection *self,
     ok = amqp_basic_cancel(self->conn, channel,
                            PyString_AS_AMQBYTES(consumer_tag));
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "basic.cancel"))
@@ -1772,6 +1904,7 @@ PyRabbitMQ_Connection_basic_consume(PyRabbitMQ_Connection *self,
 
     amqp_basic_consume_ok_t *ok;
     amqp_rpc_reply_t reply;
+    amqp_pool_t *channel_pool = NULL;
     amqp_table_t cargs = AMQP_EMPTY_TABLE;
 
     if (PyRabbitMQ_Not_Connected(self))
@@ -1784,7 +1917,12 @@ PyRabbitMQ_Connection_basic_consume(PyRabbitMQ_Connection *self,
     if ((queue = Maybe_Unicode(queue)) == NULL) goto bail;
     if ((consumer_tag = Maybe_Unicode(consumer_tag)) == NULL) goto bail;
 
-    cargs = PyDict_ToAMQTable(self->conn, arguments);
+    channel_pool = amqp_get_or_create_channel_pool(self->conn, (amqp_channel_t)channel);
+    if (channel_pool == NULL) {
+        PyErr_NoMemory();
+        goto bail;
+    }
+    cargs = PyDict_ToAMQTable(self->conn, arguments, channel_pool);
     if (PyErr_Occurred()) goto bail;
 
     Py_BEGIN_ALLOW_THREADS;
@@ -1859,7 +1997,7 @@ PyRabbitMQ_Connection_flow(PyRabbitMQ_Connection *self,
     Py_BEGIN_ALLOW_THREADS;
     ok = amqp_channel_flow(self->conn, channel, (amqp_boolean_t)active);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "channel.flow"))
@@ -1892,7 +2030,7 @@ PyRabbitMQ_Connection_basic_recover(PyRabbitMQ_Connection *self,
     Py_BEGIN_ALLOW_THREADS;
     ok = amqp_basic_recover(self->conn, channel, (amqp_boolean_t)requeue);
     reply = amqp_get_rpc_reply(self->conn);
-    amqp_maybe_release_buffers(self->conn);
+    amqp_maybe_release_buffers_on_channel(self->conn, channel);
     Py_END_ALLOW_THREADS;
 
     if (PyRabbitMQ_HandleAMQError(self, channel, reply, "basic.recover"))
