@@ -3,6 +3,7 @@ import socket
 
 import _librabbitmq
 
+from amqp.protocol import queue_declare_ok_t
 from array import array
 
 __version__ = _librabbitmq.__version__
@@ -15,7 +16,7 @@ ConnectionError = _librabbitmq.ConnectionError
 ChannelError = _librabbitmq.ChannelError
 
 
-__version__ = '1.0.3'
+__version__ = '1.5.2'
 __all__ = ['Connection', 'Message', 'ConnectionError', 'ChannelError']
 
 
@@ -124,16 +125,19 @@ class Channel(object):
         return self.connection._exchange_declare(self.channel_id,
                 exchange, type, passive, durable, auto_delete, arguments or {})
 
-    def exchange_delete(self, exchange='', if_unused=False):
-        return self.connection._exchange_delete(self.channel_id,
-                exchange, if_unused)
+    def exchange_delete(self, exchange='', if_unused=False, nowait=False):
+        return self.connection._exchange_delete(
+                self.channel_id, exchange, if_unused)
 
     def queue_declare(self, queue='', passive=False, durable=False,
-            exclusive=False, auto_delete=False, arguments=None,
-            nowait=False):
-        return self.connection._queue_declare(self.channel_id,
-                queue, passive, durable, exclusive, auto_delete,
-                arguments or {})
+                      exclusive=False, auto_delete=False, arguments=None,
+                      nowait=False):
+        return queue_declare_ok_t(
+            *self.connection._queue_declare(
+                self.channel_id, queue, passive, durable,
+                exclusive, auto_delete, arguments or {},
+            )
+        )
 
     def queue_bind(self, queue='', exchange='', routing_key='',
             arguments=None, nowait=False):
